unit Unit1;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, sqlitePassUtils;

type
  TForm1 = class(TForm)
    StrTest: TButton;
    WideStrTest: TButton;
    ZStrTest: TButton;
    Iterations: TEdit;
    Label1: TLabel;
    StrTime: TStaticText;
    ZStrTime: TStaticText;
    WideStrTime: TStaticText;
    RawStrTest: TButton;
    RawStrTime: TStaticText;
    Memo: TMemo;
    StrTestTL: TButton;
    StrTimeTL: TStaticText;
    StrTestSplitePassStrList: TButton;
    StrTimeSplitePassStrList: TStaticText;
    procedure StrTestClick(Sender: TObject);
    procedure WideStrTestClick(Sender: TObject);
    procedure ZStrTestClick(Sender: TObject);
    procedure RawStrTestClick(Sender: TObject);
    procedure StrTimeTLClick(Sender: TObject);
    procedure StrTestSplitePassStrListClick(Sender: TObject);
  private
    { Dclarations prives }
  public
    { Dclarations publiques }
  end;

var
  Form1: TForm1;
  ElapsedTime: Double;
  StartTime:Integer;
  StopTime: Integer;
  IterationsCount: Integer;
//  s : TSqlitePassStringList;
  
implementation

{$R *.DFM}

{ Test with Dyn Array if AnsiString
  (+)
  This is Almost the fastest way but need more memory for small tables...
  the memory is first allocated to hold pointers with a call to
  SetLength(S, IterationsCount);
  This doesn't impact to much memory because only pointers, not string content
  is allocated.
  (-)
  in order to limit waste of memory you should know how many records you have to store
  }
procedure TForm1.StrTestClick(Sender: TObject);
Const
//V : PAnsiChar = 'abcddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd';// + IntToStr(i);
V : PAnsiChar = 'a';// + IntToStr(i);
var
i: Integer;
FCapacity: Integer;
//S: Array [0..1000000] of AnsiString; don't work needs to much memory
S: Array of AnsiString;

procedure Grow;
var
  Delta: Integer;
begin
  if FCapacity > 64 then Delta := FCapacity div 4 else
    if FCapacity > 8 then Delta := 16 else
      Delta := 4;
  Inc(FCapacity, Delta);
  SetLength(S, FCapacity);
end;

begin
IterationsCount := StrToInt(Iterations.Text);
StrTime.Caption := '';
StartTime:=Integer(GetTickCount);
FCapacity := 4;
//SetLength(S, IterationsCount);
SetLength(S, MaxListSize);

For i := 0 to Pred(IterationsCount) do
    begin
    if i = Pred(FCapacity) then Grow;
    S[i] := StrPas(V);
    end;
SetLength(S,0);
StopTime:=Integer(GetTickCount);
ElapsedTime := (StopTime - StartTime) / 1000;
StrTime.Caption := FloatToStrF(ElapsedTime, ffNumber, 8, 4);
end;

{ Test with TStringList
  (+)
  Memory is used with more efficiency
  You don't have to  know how many records you have to store
  (-)
  a little bit slower than Dyn Array
  }

procedure TForm1.StrTimeTLClick(Sender: TObject);
Const
//V: AnsiString = 'abcddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd';
V : AnsiString = 'a';
var
i: Integer;
Tl: TStringList;
begin
IterationsCount := StrToInt(Iterations.Text);
StrTimeTL.Caption := '';
StartTime:=Integer(GetTickCount);
TL := TStringList.Create;
TL.Capacity := 8;
For i := 0 to Pred(IterationsCount) do
    begin
    TL.Add(V);
    end;
TL.Free;
StopTime:=Integer(GetTickCount);
ElapsedTime := (StopTime - StartTime) / 1000;
StrTimeTL.Caption := FloatToStrF(ElapsedTime, ffNumber, 8, 4);
end;

{ Test with TSqlitePassStringList
  (+)
  Memory is used with more efficiency
  You don't have to  know how many records you have to store
  A little faster than TStringList
  }
procedure TForm1.StrTestSplitePassStrListClick(Sender: TObject);
Const
//V: AnsiString = 'abcddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd';
V : AnsiString = 'a';
var
i: Integer;
Tl: TSqlitePassStringList;
begin
IterationsCount := StrToInt(Iterations.Text);
StrTimeSplitePassStrList.Caption := '';
StartTime:=Integer(GetTickCount);
TL := TSqlitePassStringList.Create;
TL.Capacity := 8;
For i := 0 to Pred(IterationsCount) do
    begin
    TL.Add(V);
    end;
TL.Free;
StopTime:=Integer(GetTickCount);
ElapsedTime := (StopTime - StartTime) / 1000;
StrTimeSplitePassStrList.Caption := FloatToStrF(ElapsedTime, ffNumber, 8, 4);
end;

{ Test with Dyn Array of WideString
  (+)
  Nothing (under d4) to be check under D9 and unicodeStrings
  D9 = Very, very slow with WideString
  D9 = Very fast with String = unicodeString

  (-)
  Same as AnsiStrings but very very slow...
  in order to limit waste of memory you should know how many records you have to store
  }

procedure TForm1.WideStrTestClick(Sender: TObject);
const
//V: WideString = 'abcddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd';// + IntToStr(i);
V: WideString = 'a';
var
i: Integer;
//S: Array of WideString;
S: Array of UnicodeString;
begin
IterationsCount := StrToInt(Iterations.Text);
WideStrTime.Caption := '';
StartTime:=Integer(GetTickCount);

SetLength(S, IterationsCount);
For i := 0 to Pred(IterationsCount) do
    begin
     S[i] := StrPas(PChar(V));
    end;
SetLength(S,0);

StopTime:=Integer(GetTickCount);
ElapsedTime := (StopTime - StartTime) / 1000;
WideStrTime.Caption := FloatToStrF(ElapsedTime, ffNumber, 8, 4);
end;

{ Test with Dyn Array of PChar and AZT
  (+)
  This is the fastest way.
  Same as
  (-)
  in order to limit waste of memory you should know how many records you have to store
  Less handy to manage (Dispose and String Conversions ? Check GetFieldData.
  }
procedure TForm1.ZStrTestClick(Sender: TObject);
Const
V: AnsiString = 'abcddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd';// + IntToStr(i);

var
i: Integer;
S: Array of PAnsiChar;
begin
IterationsCount := StrToInt(Iterations.Text);
ZStrTime.Caption := '';
StartTime:=Integer(GetTickCount);

SetLength(S, IterationsCount);
For i := 0 to Pred(IterationsCount) do
    begin
    S[i] := StrNew(PAnsiChar(V));
    end;

For i := 0 to Pred(IterationsCount) do
    begin
    StrDispose(S[i]);
    end;

SetLength(S,0);

StopTime:=Integer(GetTickCount);
ElapsedTime := (StopTime - StartTime) / 1000;
ZStrTime.Caption := FloatToStrF(ElapsedTime, ffNumber, 8, 4);
end;

procedure TForm1.RawStrTestClick(Sender: TObject);
//Const
//V: AnsiString = 'abcddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd';
var
i: Integer;
S: Array of PAnsiChar;//PAnsiChar;
//V: AnsiString;
V: PAnsiChar;

begin
IterationsCount := StrToInt(Iterations.Text);
RawStrTime.Caption := '';
StartTime:=Integer(GetTickCount);
V := 'abcddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddddd';
SetLength(S, IterationsCount);
For i := 0 to Pred(IterationsCount) do
    begin
//    V:= 'abc' + IntToStr(i);
    GetMem(S[i],Length(V)+1);
//    Move(V,S[i],Length(V));
//    S[I]+Length(v
//    StrCopy(V,S[i]);
    FreeMem(S[i]);
//    Memo.Lines.Add(AnsiString(S[i]));

    end;

(*For i := 0 to Pred(IterationsCount) do
    begin
    Memo.Lines.Add(AnsiString(S[i]));

    FreeMem(S[i]);
    end;

SetLength(S,0);*)

StopTime:=Integer(GetTickCount);
ElapsedTime := (StopTime - StartTime) / 1000;
RawStrTime.Caption := FloatToStrF(ElapsedTime, ffNumber, 8, 4);

end;



end.
