{ This library is free software; you can redistribute it and/or modify it
  under the terms of the GNU Library General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU Library General Public License
  for more details.

  You should have received a copy of the GNU Library General Public License
  along with this library; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

  ---------------------------------------------------------------------------

    Author : Luc DAVID Email: luckylazarus@free.fr
    2007 - 2009
    Last update : 19/04/2008

  --------------------------------------------------------------------------- }

unit SqlitePassFiltersDialog;
{$i ..\..\Sources\SqlitePassDbo.inc}

interface

uses
 {$IFDEF FPC}
  LResources,
 {$ELSE}
  Windows, Messages,
 {$ENDIF}
  SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, Buttons, db, Math,
  SqlitePassDbo, CheckLst, ComCtrls;

type
  TSqlitePassFilterDialog = class(TForm)
    PanelIndexApplyChanges: TPanel;
    SbCancel: TSpeedButton;
    SbOk: TSpeedButton;
    PageControl: TPageControl;
    tsCustomFilters: TTabSheet;
    tsSQLFilters: TTabSheet;
    PanelIndexesToolBar: TPanel;
    Bevel1: TBevel;
    LabelFiltersDefinitions: TLabel;
    PanelSetFilterText: TPanel;
    PanelIndexMain: TPanel;
    PanelAvailableIndexes: TPanel;
    PanelCaptionAvailableIFields: TPanel;
    Image5: TImage;
    CheckListBoxFields: TCheckListBox;
    PanelSQL: TPanel;
    SbSelectValueByList1: TSpeedButton;
    SbSelectValueByList2: TSpeedButton;
    Bevel3: TBevel;
    SpeedButton4: TSpeedButton;
    SbClearFilter: TSpeedButton;
    LabelDataFormat: TLabel;
    PanelIndexCreateStmt: TPanel;
    Label21: TLabel;
    Image4: TImage;
    ComboBoxFilterOperator1: TComboBox;
    RbAnd: TRadioButton;
    RbOr: TRadioButton;
    EditFilterValue1: TEdit;
    ComboBoxFilterOperator2: TComboBox;
    EditFilterValue2: TEdit;
    Edit1: TEdit;
    Panel1: TPanel;
    SpeedButton3: TSpeedButton;
    SpeedButton5: TSpeedButton;
    SpeedButton6: TSpeedButton;
    tsCountLimitFilter: TTabSheet;
    Bevel2: TBevel;
    Label1: TLabel;
    LabelLowerLimit: TLabel;
    Label2: TLabel;
    EditLowerLimit: TEdit;
    EditUpperLimit: TEdit;
    SbClearFilterRanges: TSpeedButton;
    Memo1: TMemo;
    Bevel4: TBevel;
    Memo2: TMemo;
    Panel2: TPanel;
    Bevel5: TBevel;
    LabelFilterText: TLabel;
    Panel4: TPanel;
    Splitter1: TSplitter;
    Panel5: TPanel;
    ListBoxFields: TListBox;
    Panel6: TPanel;
    Image1: TImage;
    PanelUseQuotes: TPanel;
    CheckBoxUseQuotes: TCheckBox;
    Panel7: TPanel;
    MemoSQLFilterStmt: TMemo;
    Panel8: TPanel;
    Label6: TLabel;
    Image2: TImage;
    SpeedButton2: TSpeedButton;
    SbDisplayFilterStmt: TSpeedButton;
    tsFilterStmt: TTabSheet;
    MemoPreviousFilterStmt: TMemo;
    Label3: TLabel;
    Bevel6: TBevel;
    SbCloseTsFilterStmt: TSpeedButton;
    MemoNewFilterStmt: TMemo;
    Label4: TLabel;
    Label5: TLabel;
    procedure SbOkClick(Sender: TObject);
    procedure SbCancelClick(Sender: TObject);
    procedure SbClearAllClick(Sender: TObject);
    procedure SbSelectValueByList1Click(Sender: TObject);
    procedure ComboBoxFilterOperator1Change(Sender: TObject);
    procedure ComboBoxFilterOperator2Change(Sender: TObject);
    procedure SbClearFilterClick(Sender: TObject);
    procedure CheckListBoxFieldsClick(Sender: TObject);
    procedure SbDisableAllClick(Sender: TObject);
    procedure SbSelectValueByList2Click(Sender: TObject);
    procedure CheckListBoxFieldsDblClick(Sender: TObject);
    procedure SbEnableAllClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SbDisplayFilterStmtClick(Sender: TObject);
    procedure SbCloseTsFilterStmtClick(Sender: TObject);
    procedure tsFilterStmtShow(Sender: TObject);
    procedure SbClearFilterRangesClick(Sender: TObject);
    procedure ListBoxFieldsDblClick(Sender: TObject);
  private
    FFilterText: String;
    PreviousSelectedFieldIndex: Integer;
    MyDb: TSqlitePassDatabase;
    MyDataset: TSqlitePassDataset;
    procedure DisplayTabSheets(Visible: Boolean);
    procedure RefreshFieldsList;
    procedure InitFiltersFromDatasetFilters;
    procedure GetFilterValues(Edit: TEdit);
    function BuildFilterStmt: String;
    procedure DisplayFilter(Index: Integer);
    procedure StoreFilter(Index: Integer);
    function ComparisonOpAsText(CompOp: TSqlitePassFilterComparisonOperators): String;
    function LogicalOpAsText(LogOp: TSqlitePassFilterLogicalOperators): String;
  public
    constructor Create(AOwner: TComponent; Dataset: TSqlitePassDataset); reintroduce;
    Property FilterText: String Read FFilterText;
  end;

 Filter = Record
    FieldName: String;
    LogOperator: TSqlitePassFilterLogicalOperators;
    CompOperator: TSqlitePassFilterComparisonOperators;
    Value: String;
    end;

var
  SqlitePassFilterDialog: TSqlitePassFilterDialog;
  Filters: Array of Array of Filter;
implementation

uses SqlitePassFilterValues;

{$IFNDEF FPC}
  {$R *.DFM}
{$ENDIF}


constructor TSqlitePassFilterDialog.Create(AOwner: TComponent; Dataset: TSqlitePassDataset);
var
i, FirstActiveFilter: Integer;
begin
inherited Create(AOwner);
FFilterText := '';
ListBoxFields.Clear;

if Dataset = nil
   then abort
   else begin
        MyDataset := Dataset;
        Caption := 'Filter Editor for ' + MyDataset.DatasetName;
        MyDb := MyDataset.Database;
        RefreshFieldsList;
        InitFiltersFromDatasetFilters;
        if CheckListBoxFields.Items.Count > 0 then
           begin
           FirstActiveFilter := 0;
           for i := 0 to Pred(CheckListBoxFields.Items.Count) do
               if CheckListBoxFields.Checked[i] then
                  begin
                  FirstActiveFilter := i;
                  Break;
                  end;
           CheckListBoxFields.ItemIndex := FirstActiveFilter;
           PreviousSelectedFieldIndex := -1;
           CheckListBoxFieldsClick(Self);
           DisplayTabSheets(True);
           end;
        end;
end;


{ Retrieve data on fields and fill the listbox with it }
procedure TSqlitePassFilterDialog.RefreshFieldsList;
var
i: integer;
Field: TField;

begin
{ We fill the listboxes with fields names }
ListBoxFields.Clear;
CheckListBoxFields.Clear;
CheckListBoxFields.Sorted := False;

For i := 0 to Pred(MyDataset.Fields.count) do
  begin
  { Listbox }
  ListBoxFields.items.Add(MyDataset.Fields[i].FieldName);
  { Check Listbox }
  Field := MyDataset.Fields[i];
  CheckListBoxFields.Items.Add(Field.FieldName);
  CheckListBoxFields.Checked[i] := MyDataset.Filters.FilterByField(Field).Filtered;
  end;

CheckListBoxFields.Sorted := True;
{ Ranges }
EditLowerLimit.Text := IntToStr(MyDataset.FilterRecordLowerLimit);
EditUpperLimit.Text := IntToStr(MyDataset.FilterRecordUpperLimit);
end;

{ Fills the filters from Dataset.Filters }
procedure TSqlitePassFilterDialog.InitFiltersFromDatasetFilters;
var
i, j, FilterIndex: Integer;
QuoteChar: String;
FieldFilter: TSqlitePassFieldFilter;
begin
SetLength(Filters, MyDataset.Filters.Count);

For i := 0 to Pred(MyDataset.Filters.Count) do
  begin
  FieldFilter := MyDataset.Filters[i];
  FilterIndex := CheckListBoxFields.Items.IndexOf(FieldFilter.Field.FieldName);
  if FieldFilter.Count < 2
     then SetLength(Filters[FilterIndex], 2)
     else SetLength(Filters[FilterIndex], FieldFilter.Count);

  For j := 0 to Pred(FieldFilter.Count) do
    begin
    QuoteChar := '';
    With Filters[FilterIndex, j] do
         begin
         FieldName := FieldFilter.Field.FieldName;
         LogOperator  := FieldFilter[j].LogicalOperator;
         CompOperator := FieldFilter[j].ComparisonOperator;
         Case FieldFilter.Field.Datatype of
              ftString, ftFixedChar, ftWideString: QuoteChar := '"';
              ftDateTime, ftDate, ftTime: QuoteChar := '#';
              end;
         Value := QuoteChar + FieldFilter[j].Value + QuoteChar;
         end;
    end;     
  end;
end;

procedure TSqlitePassFilterDialog.SbCancelClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

{ Clears ALL the filters }
procedure TSqlitePassFilterDialog.SbClearAllClick(Sender: TObject);
begin
SbDisableAllClick(Self);
MyDataset.Filters.ClearFilters;
SetLength(Filters, MyDataset.Filters.Count);
SbClearFilterClick(Sender);
end;

{ Clears the currently selected filter }
procedure TSqlitePassFilterDialog.SbClearFilterClick(Sender: TObject);
begin
 ComboBoxFilterOperator1.ItemIndex := 0;
 EditFilterValue1.Text := '';
 ComboBoxFilterOperator2.ItemIndex := 0;
 EditFilterValue2.Text := '';
end;

{ Displays the Filter content for a given FieldDef (Index = CheckListBox.ItemIndex) }
procedure TSqlitePassFilterDialog.DisplayFilter(Index: Integer);
begin
  SbClearFilterClick(Self);
  ComboBoxFilterOperator1.ItemIndex := Ord(Filters[Index, 0].CompOperator) ;
  EditFilterValue1.Text := Filters[Index, 0].Value;
  RbAnd.Checked := Filters[Index, 1].LogOperator = opAnd;
  RbOr.Checked := Filters[Index, 1].LogOperator = opOr;
  ComboBoxFilterOperator2.ItemIndex := Ord(Filters[Index, 1].CompOperator) ;
  EditFilterValue2.Text := Filters[Index, 1].Value;
end;

Procedure TSqlitePassFilterDialog.StoreFilter(Index: Integer);
begin
With Filters[Index, 0] do
     begin
     FieldName := CheckListBoxFields.Items[Index];
     LogOperator := opNone;
     if ComboBoxFilterOperator1.ItemIndex > -1
        then CompOperator := TSqlitePassFilterComparisonOperators(ComboBoxFilterOperator1.ItemIndex)
        else CompOperator := CmpUnknown;
     Value := EditFilterValue1.Text;
     end;
With Filters[Index, 1] do
     begin
     FieldName := CheckListBoxFields.Items[Index];
     if RbAnd.Checked
        then LogOperator := opAnd
        else LogOperator := opOr;
     if ComboBoxFilterOperator2.ItemIndex > -1
        then CompOperator := TSqlitePassFilterComparisonOperators(ComboBoxFilterOperator2.ItemIndex)
        else CompOperator := CmpUnknown;
     Value := EditFilterValue2.Text;
     end;
end;

function TSqlitePassFilterDialog.BuildFilterStmt: String;
var
i, j: Integer;
Field: TField;
SQLStr, FieldName: String;

begin
SQLStr := MemoSQLFilterStmt.Text;
Result := '';

{ Validate the current filter }
CheckListBoxFieldsClick(nil);

for i := 0 to High(Filters) do
  begin
  for j := 0 to High(Filters[i]) do
    if (Filters[i, j].FieldName <> '')
       and (Filters[i, j].CompOperator <> cmpUnknown)
       (*and (Filters[i, j].Value <> '' *) then
       begin
       Field := MyDataset.Fields.FieldByName(Filters[i,j].FieldName);
       if Field <> nil then
          begin
          if FieldName = Field.FieldName
             then FieldName := ''
             else FieldName := Field.FieldName;

          Result := Result
                    + FieldName
                    + LogicalOpAsText(Filters[i,j].LogOperator)
                    + ComparisonOpAsText(Filters[i,j].CompOperator)
                    + Filters[i,j].Value;
          end;
       end;

     if (Result <> '') and (Result[Length(Result)] <> ';')
        then Result :=  ';' + Result;
   end;
   Result := SQLStr + Result;   
end;


{ Add the filters defined to the Dataset.Filters }
procedure TSqlitePassFilterDialog.SbOkClick(Sender: TObject);
begin
FFilterText := BuildFilterStmt;

With MyDataset do
     begin
     Filter := FFilterText;
     FilterRecordLowerLimit := StrToIntDef(EditLowerLimit.Text,0);
     FilterRecordUpperLimit := StrToIntDef(EditUpperLimit.Text,0);
     end;

ModalResult := mrOk;
end;

procedure TSqlitePassFilterDialog.CheckListBoxFieldsClick(Sender: TObject);
var
Field: TField;
begin
{ Saves the filters settings for the previous selected field }
if PreviousSelectedFieldIndex > -1
   then StoreFilter(PreviousSelectedFieldIndex);

{ Show the filters of the new selected field }
if PreviousSelectedFieldIndex <> CheckListBoxFields.ItemIndex then
   begin
   PreviousSelectedFieldIndex := CheckListBoxFields.ItemIndex;
   DisplayFilter(PreviousSelectedFieldIndex);
   Field := MyDataset.Fields.FieldByName(CheckListBoxFields.Items[PreviousSelectedFieldIndex]);
   Case Field.Datatype of
       ftString,
       ftWideString : LabelDataFormat.Caption := 'Data format : You can use * or % as WildCard chars';
       ftDateTime : LabelDataFormat.Caption := 'Data format : #' + MyDataset.Database.DatatypeOptions.DateTimeFormat + '#';
       ftDate : LabelDataFormat.Caption := 'Data format : #' + MyDataset.Database.DatatypeOptions.DateFormat + '#';
       ftTime : LabelDataFormat.Caption := 'Data format : #' + MyDataset.Database.DatatypeOptions.TimeFormat + '#';
       else     LabelDataFormat.Caption := '';
       end;
   end;
end;

procedure TSqlitePassFilterDialog.SbDisableAllClick(Sender: TObject);
var
i: Integer;
begin
 For i := 0 to Pred(CheckListBoxFields.Items.Count)
     do CheckListBoxFields.Checked[i] := False;
end;

procedure TSqlitePassFilterDialog.SbEnableAllClick(Sender: TObject);
var
i: Integer;
begin
 For i := 0 to Pred(CheckListBoxFields.Items.Count)
     do CheckListBoxFields.Checked[i] := True;
end;

procedure TSqlitePassFilterDialog.CheckListBoxFieldsDblClick(
  Sender: TObject);
begin
 CheckListBoxFields.Checked[CheckListBoxFields.ItemIndex] := Not CheckListBoxFields.Checked[CheckListBoxFields.ItemIndex];
end;

{ Gets one or more filter value from Value list Dialog }

procedure TSqlitePassFilterDialog.ComboBoxFilterOperator1Change(Sender: TObject);
begin
 SbSelectValueByList1.Enabled := (ComboBoxFilterOperator1.Text = '=')
                              or (ComboBoxFilterOperator1.Text = '<>')
                              or (ComboBoxFilterOperator1.Text = 'Like');
end;

procedure TSqlitePassFilterDialog.ComboBoxFilterOperator2Change(Sender: TObject);
begin
 if Not (RbAnd.Checked or RbOr.Checked)
    then if ComboBoxFilterOperator1.Text = '='
            then RbOr.Checked := True
            else RbAnd.Checked := True;
 SbSelectValueByList2.Enabled := (ComboBoxFilterOperator2.Text = '=')
                              or (ComboBoxFilterOperator1.Text = '<>')
                              or (ComboBoxFilterOperator2.Text = 'Like');
end;

procedure TSqlitePassFilterDialog.SbSelectValueByList1Click(Sender: TObject);
begin
  GetFilterValues(EditFilterValue1);
end;

procedure TSqlitePassFilterDialog.SbSelectValueByList2Click(
  Sender: TObject);
begin
 GetFilterValues(EditFilterValue2);
end;

procedure TSqlitePassFilterDialog.GetFilterValues(Edit: TEdit);
var
ValueDlg: TSqlitePassFilterValuesDlg;
CurrentField: TField;
begin
if CheckListBoxFields.ItemIndex > -1 then
   begin
   CurrentField := MyDataset.FindField(CheckListBoxFields.Items[CheckListBoxFields.ItemIndex]);
   ValueDlg := TSqlitePassFilterValuesDlg.Create(Self, MyDataset, CurrentField);
   if ValueDlg.ShowModal = mrOk
      then Edit.Text := ValueDlg.FilterText;
   ValueDlg.Free;
   end;
end;

function TSqlitePassFilterDialog.ComparisonOpAsText(
  CompOp: TSqlitePassFilterComparisonOperators): String;
begin
Case CompOp of
  CmpEqual         : Result := ' = ';
  CmpNotEqual      : Result := ' <> ';
  CmpGreater       : Result := ' > ';
  CmpGreateOrEqual : Result := ' >= ';
  CmpLesserOrEqual : Result := ' <= ';
  CmpLesser        : Result := ' < ';
  CmpNull          : Result := ' Null ';
  CmpNotNull       : Result := ' Not Null ';
  else               Result := '';
  end;
end;

function TSqlitePassFilterDialog.LogicalOpAsText(
  LogOp: TSqlitePassFilterLogicalOperators): String;
begin
Case LogOp of
  opAnd : Result := ' And';
  opOr  : Result := ' Or';
  else    Result := '';
  end;
end;

procedure TSqlitePassFilterDialog.FormDestroy(Sender: TObject);
begin
SetLength(Filters,0);
end;

procedure TSqlitePassFilterDialog.SbDisplayFilterStmtClick(
  Sender: TObject);
begin
  DisplayTabSheets(False);
end;

procedure TSqlitePassFilterDialog.SbCloseTsFilterStmtClick(
  Sender: TObject);
begin
  DisplayTabSheets(True);
end;

procedure TSqlitePassFilterDialog.DisplayTabSheets(Visible: Boolean);
begin
 tsCustomFilters.TabVisible := Visible;
 tsSQLFilters.TabVisible := Visible;
 tsCountLimitFilter.TabVisible := Visible;
 tsFilterStmt.TabVisible := Not Visible;
 if tsCustomFilters.TabVisible
    then PageControl.ActivePage := tsCustomFilters
    else PageControl.ActivePage := tsFilterStmt;
 SbDisplayFilterStmt.Enabled := Visible;   
end;

procedure TSqlitePassFilterDialog.tsFilterStmtShow(Sender: TObject);
begin
 MemoPreviousFilterStmt.Text := MyDataset.Filter;
 MemoNewFilterStmt.Text := BuildFilterStmt;
end;

procedure TSqlitePassFilterDialog.SbClearFilterRangesClick(
  Sender: TObject);
begin
EditLowerLimit.Text := '0';
EditUpperLimit.Text := '0';
end;

procedure TSqlitePassFilterDialog.ListBoxFieldsDblClick(Sender: TObject);
var
SelectedField: String;
begin
SelectedField := ListBoxFields.Items[ListBoxFields.ItemIndex];
if CheckBoxUseQuotes.Checked
   then MyDataset.SQLSelectStmt.QuoteString(SelectedField);

MemoSQLFilterStmt.SetFocus;
MemoSQLFilterStmt.SelText := SelectedField;
end;

initialization
 {$IFDEF FPC}
  {$I SqlitePassFiltersDialog.lrs}
 {$ENDIF}
end.
