{**************************************************************************************************}
{                                                                                                  }
{ UTF-8 VCL                                                                                        }
{                                                                                                  }
{ The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); }
{ you may not use this file except in compliance with the License. You may obtain a copy of the    }
{ License at http://www.mozilla.org/MPL/                                                           }
{                                                                                                  }
{ Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF   }
{ ANY KIND, either express or implied. See the License for the specific language governing rights  }
{ and limitations under the License.                                                               }
{                                                                                                  }
{ The Original Code is UTF8VCLCommDlg.pas                                                                 }
{                                                                                                  }
{ The Initial Developer of the Original Code is Bart van der Werf (bwerf, bluelive)                }
{ Portions created by Bart van der Werf are Copyright (C) Bart van der Werf. All rights reserved.  }
{                                                                                                  }
{ Contributor(s):                                                                                  }
{   Andreas Hausladen (ahuser)                                                                     }
{   Robert Marquardt (rom, marquardt)                                                              }
{                                                                                                  }
{**************************************************************************************************}

// Last modified: $Date: 2007/08/17 14:39:24 $

unit UTF8VCLCommDlg;

interface

uses
  Windows, CommDlg, Messages, SysUtils, Classes, UTF8VCLUtils, UTF8VCLMessages;

function GetOpenFileNameAIntercept(var OpenFileA: TOpenFilenameA): Bool; stdcall;
function GetSaveFileNameAIntercept(var OpenFileA: TOpenFilenameA): Bool; stdcall;
function GetFileTitleAIntercept(FileName: PAnsiChar; Title: PAnsiChar; TitleSize: Word): Smallint; stdcall;
function ChooseColorAIntercept(var CC: TChooseColorA): Bool; stdcall;

implementation


function ConvertFilterStringFrom(const S: PChar): PChar;
var
  I, L: Integer;
  P: Boolean;
begin
  L := 0;
  P := False;
  while Assigned(S) and not (P and (S[L] = #0)) do
  begin
    P := S[L] = #0;
    Inc(L);
  end;
  if L = 0 then
  begin
    Result := nil;
    Exit;
  end
  else
  begin
    Inc(L);
    GetMem(Result, L);
    for I := 0 to L - 2 do
    begin
      if S[I] = #0 then
        Result[I] := #1
      else
        Result[I] := S[I];
    end;
    Result[L-1] := #0;
  end;
end;

procedure ConvertFilterStringTo(S: PWideChar);
var
  I: Integer;
begin
  I := 0;
  while Assigned(S) and (S[I] <> #0) do
  begin
    if S[I] = #1 then
      S[I] := #0;
    Inc(I);
  end;
end;

function GetOpenFileNameAIntercept(var OpenFileA: TOpenFilenameA): Bool; stdcall;
var
  OpenFileW: TOpenFilenameW;
  X: PChar;
  S: PChar;
  L: Integer;
begin
  Move(OpenFileA, OpenFileW, SizeOf(TOpenFilenameA));
  with OpenFileW do
  begin
    lpstrFilter := nil;
    lpstrInitialDir := nil;
    lpstrTitle := nil;
    lpstrDefExt := nil;
    lpstrCustomFilter := nil;
    lpstrFile := nil;
    lpstrFileTitle := nil;
    lpTemplateName := nil;
  end;
  X := nil;
  S := nil;

  try
    S := ConvertFilterStringFrom(OpenFileA.lpstrFilter);
    OpenFileW.lpstrFilter := ConvertFromUTF8(S);
    ConvertFilterStringTo(OpenFileW.lpstrFilter);

    OpenFileW.lpstrInitialDir := ConvertFromUTF8(OpenFileA.lpstrInitialDir);
    OpenFileW.lpstrTitle := ConvertFromUTF8(OpenFileA.lpstrTitle);
    OpenFileW.lpstrDefExt := ConvertFromUTF8(OpenFileA.lpstrDefExt);
    OpenFileW.lpTemplateName := ConvertFromUTF8(OpenFileA.lpTemplateName);

    OpenFileW.lpstrFile := ConvertFromUTF8(OpenFileA.lpstrFile);
    if OpenFileA.nMaxFile <> 0 then
      ReallocMem(OpenFileW.lpstrFile, OpenFileA.nMaxFile*2 + 4);
    OpenFileW.lpstrCustomFilter := ConvertFromUTF8(OpenFileA.lpstrCustomFilter);
    if OpenFileA.nMaxCustFilter <> 0 then
      ReallocMem(OpenFileW.lpstrCustomFilter, OpenFileA.nMaxCustFilter*2 + 4);
    OpenFileW.lpstrFileTitle := ConvertFromUTF8(OpenFileA.lpstrFileTitle);
    if OpenFileA.nMaxFileTitle <> 0 then
      ReallocMem(OpenFileW.lpstrFileTitle, OpenFileA.nMaxFileTitle*2 + 4);

    OpenFileW.lpfnHook := WndProcConvertW2U(Pointer(@OpenFileA.lpfnHook),
      Pointer(@StdWndProcInterceptU2W), Pointer(@StdWndProcInterceptW2U));

    Result := GetOpenFileNameW(OpenFileW);

    X := ConvertToUTF8(OpenFileW.lpstrFile, -1, L);
    CopyMemory(OpenFileA.lpstrFile, X, L+1);
    FreeMem(X);
    X := ConvertToUTF8(OpenFileW.lpstrCustomFilter, -1, L);
    CopyMemory(OpenFileA.lpstrCustomFilter, X, L+1);
    FreeMem(X);
    X := ConvertToUTF8(OpenFileW.lpstrFileTitle, -1, L);
    CopyMemory(OpenFileA.lpstrFileTitle, X, L+1);
  finally
    FreeMem(S);
    FreeMem(X);
    with OpenFileW do
    begin
      FreeMem(lpstrFilter);
      FreeMem(lpstrInitialDir);
      FreeMem(lpstrTitle);
      FreeMem(lpstrDefExt);
      FreeMem(lpstrFile);
      FreeMem(lpstrCustomFilter);
      FreeMem(lpstrFileTitle);
      FreeMem(lpTemplateName);
    end;
  end;
end;

// /!\ copypaste form getopenfilenameaintercept
function GetSaveFileNameAIntercept(var OpenFileA: TOpenFilenameA): Bool; stdcall;
var
  OpenFileW: TOpenFilenameW;
  X: PChar;
  S: PChar;
  L: Integer;
begin
  Move(OpenFileA, OpenFileW, SizeOf(TOpenFilenameA));
  with OpenFileW do
  begin
    lpstrFilter := nil;
    lpstrInitialDir := nil;
    lpstrTitle := nil;
    lpstrDefExt := nil;
    lpstrCustomFilter := nil;
    lpstrFile := nil;
    lpstrFileTitle := nil;
    lpTemplateName := nil;
  end;
  X := nil;
  S := nil;

  try
    S := ConvertFilterStringFrom(OpenFileA.lpstrFilter);
    OpenFileW.lpstrFilter := ConvertFromUTF8(S);
    ConvertFilterStringTo(OpenFileW.lpstrFilter);

    OpenFileW.lpstrInitialDir := ConvertFromUTF8(OpenFileA.lpstrInitialDir);
    OpenFileW.lpstrTitle := ConvertFromUTF8(OpenFileA.lpstrTitle);
    OpenFileW.lpstrDefExt := ConvertFromUTF8(OpenFileA.lpstrDefExt);
    OpenFileW.lpTemplateName := ConvertFromUTF8(OpenFileA.lpTemplateName);

    OpenFileW.lpstrFile := ConvertFromUTF8(OpenFileA.lpstrFile);
    if OpenFileA.nMaxFile <> 0 then
      ReallocMem(OpenFileW.lpstrFile, OpenFileA.nMaxFile*2 + 4);
    OpenFileW.lpstrCustomFilter := ConvertFromUTF8(OpenFileA.lpstrCustomFilter);
    if OpenFileA.nMaxCustFilter <> 0 then
      ReallocMem(OpenFileW.lpstrCustomFilter, OpenFileA.nMaxCustFilter*2 + 4);
    OpenFileW.lpstrFileTitle := ConvertFromUTF8(OpenFileA.lpstrFileTitle);
    if OpenFileA.nMaxFileTitle <> 0 then
      ReallocMem(OpenFileW.lpstrFileTitle, OpenFileA.nMaxFileTitle*2 + 4);

    OpenFileW.lpfnHook := WndProcConvertW2U(Pointer(@OpenFileA.lpfnHook),
      Pointer(@StdWndProcInterceptU2W), Pointer(@StdWndProcInterceptW2U));

    Result := GetSaveFileNameW(OpenFileW);

    X := ConvertToUTF8(OpenFileW.lpstrFile, -1, L);
    CopyMemory(OpenFileA.lpstrFile, X, L+1);
    FreeMem(X);
    X := ConvertToUTF8(OpenFileW.lpstrCustomFilter, -1, L);
    CopyMemory(OpenFileA.lpstrCustomFilter, X, L+1);
    FreeMem(X);
    X := ConvertToUTF8(OpenFileW.lpstrFileTitle, -1, L);
    CopyMemory(OpenFileA.lpstrFileTitle, X, L+1);
  finally
    FreeMem(S);
    FreeMem(X);
    with OpenFileW do
    begin
      FreeMem(lpstrFilter);
      FreeMem(lpstrInitialDir);
      FreeMem(lpstrTitle);
      FreeMem(lpstrDefExt);
      FreeMem(lpstrFile);
      FreeMem(lpstrCustomFilter);
      FreeMem(lpstrFileTitle);
      FreeMem(lpTemplateName);
    end;
  end;
end;

function GetFileTitleAIntercept(FileName: PAnsiChar; Title: PAnsiChar; TitleSize: Word): Smallint; stdcall;
var
  WideFileName, TitleBuf: PWideChar;
  X: PChar;
begin
  WideFileName := nil;
  TitleBuf := nil;
  try
    TitleBuf := GetMemory(TitleSize * SizeOf(WideChar)); //utf16 won't ever take more then 2* as much space for a string then utf8
    WideFileName := ConvertFromUTF8(FileName);
    Result := GetFileTitleW(WideFileName, TitleBuf, TitleSize);
    if Result > 0 then
      Result := Result * 4; // ensure enough size is available
    if Result = 0 then
    begin
      X := ConvertToUTF8(TitleBuf);
      try
        CopyMemory(Title, X, TitleSize);
      finally
        FreeMem(X);
      end;
    end;
  finally
    FreeMem(TitleBuf);
    FreeMem(WideFileName);
  end;
end;

function ChooseColorAIntercept(var CC: TChooseColorA): Bool; stdcall;
var
  CCW: TChooseColorW;
begin
  Move(CC, CCW, SizeOf(TChooseColorA));
  with CCW do
  begin
    lpfnHook := nil;
    lpTemplateName := nil;
  end;
  try
    if CC.Flags and CC_ENABLETEMPLATE <> 0 then
      CCW.lpTemplateName := ConvertFromUTF8(CC.lpTemplateName);
    CCW.lpfnHook := WndProcConvertW2U(Pointer(@CC.lpfnHook),
      Pointer(@StdWndProcInterceptU2W), Pointer(@StdWndProcInterceptW2U));
    Result := ChooseColorW(CCW);
    CC.rgbResult := CCW.rgbResult;
  finally
    FreeMem(CCW.lpTemplateName);
  end;
end;

{
function FindTextAIntercept(var FindReplaceA: TFindReplaceA): HWND; stdcall;
function ReplaceTextAIntercept(var FindReplace: TFindReplaceA): HWND; stdcall;
function ChooseFontAIntercept(var ChooseFont: TChooseFontA): Bool; stdcall;
function PrintDlgAIntercept(var PrintDlg: TPrintDlgA): Bool; stdcall;
function PageSetupDlgAIntercept(var PgSetupDialog: TPageSetupDlgA): BOOL; stdcall;
}
end.
