{**************************************************************************************************}
{                                                                                                  }
{ UTF-8 VCL                                                                                        }
{                                                                                                  }
{ The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); }
{ you may not use this file except in compliance with the License. You may obtain a copy of the    }
{ License at http://www.mozilla.org/MPL/                                                           }
{                                                                                                  }
{ Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF   }
{ ANY KIND, either express or implied. See the License for the specific language governing rights  }
{ and limitations under the License.                                                               }
{                                                                                                  }
{ The Original Code is UTF8VCLMessages.pas                                                         }
{                                                                                                  }
{ The Initial Developer of the Original Code is Bart van der Werf (bwerf, bluelive)                }
{ Portions created by Bart van der Werf are Copyright (C) Bart van der Werf. All rights reserved.  }
{                                                                                                  }
{ Contributor(s):                                                                                  }
{   Andreas Hausladen (ahuser)                                                                     }
{   Robert Marquardt (rom, marquardt)                                                              }
{                                                                                                  }
{**************************************************************************************************}

// Last modified: $Date: 2007/03/31 17:03:48 $

unit UTF8VCLMessages;

interface

{$IFDEF CONDITIONALEXPRESSIONS}
 {$IF CompilerVersion >= 15.0}
  {$WARN UNSAFE_TYPE OFF}
  {$WARN UNSAFE_CODE OFF}
  {$WARN UNSAFE_CAST OFF}
 {$IFEND}
{$ENDIF}

uses
  Windows, UTF8VCLUtils;

function StdWndProcInterceptU2W(Original: TStdWndProcDef; hWnd: HWND; Msg, WParam: Longint; LParam: Longint): Longint;
function StdWndProcInterceptW2U(Original: TStdWndProcDef; hWnd: HWND; Msg, WParam: Longint; LParam: Longint): Longint;

procedure MessageW2UBefore(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  out OutWParam: WPARAM; out OutLParam: LPARAM; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef);
function MessageW2UAfter(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  TempWParam: WPARAM; TempLParam: LPARAM; TempResult: LRESULT; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef): LRESULT;
procedure MessageU2WBefore(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  out OutWParam: WPARAM; out OutLParam: LPARAM; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef);
function MessageU2WAfter(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  TempWParam: WPARAM; TempLParam: LPARAM; TempResult: LRESULT; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef): LRESULT;

implementation

uses
  Messages, SysUtils;

const
  UnusedWParam = WPARAM(0);
  UnusedLParam = LPARAM(0);

function StdWndProcInterceptW2U(Original: TStdWndProcDef; hWnd: HWND; Msg, WParam: Longint; LParam: Longint): Longint;
var
  TempWParam, TempLParam: Longint;
begin
  if SpecialWndProc(Cardinal(@Original)) then
    Result := CallWindowProcW(@Original, hWnd, Msg, WParam, LParam)
  else
  if (@DefWindowProcA = @Original) or (@DefWindowProc = @Original) then
    Result := CallWindowProcW(@DefWindowProcW, hWnd, Msg, WParam, LParam)
  else
  if (@DefMDIChildProcA = @Original) or (@DefMDIChildProc = @Original) then
    Result := CallWindowProcW(@DefMDIChildProcW, hWnd, Msg, WParam, LParam)
  else
  if (@DefDlgProcA = @Original) or (@DefDlgProc = @Original) then
    Result := CallWindowProcW(@DefDlgProcW, hWnd, Msg, WParam, LParam)
  else
  begin
    MessageW2UBefore(hWnd, 0, Msg, WParam, LParam, TempWParam, TempLParam, Original, nil);
    Result := Original(hWnd, Msg, TempWParam, TempLParam);
    Result := MessageW2UAfter(hWnd, 0, Msg, WParam, LParam, TempWParam, TempLParam, Result, Original, nil);
  end;
end;

function StdWndProcInterceptU2W(Original: TStdWndProcDef; hWnd: HWND; Msg, WParam: Longint; LParam: Longint): Longint;
var
  TempWParam, TempLParam: Longint;
begin
  MessageU2WBefore(hWnd, 0, Msg, WParam, LParam, TempWParam, TempLParam, Pointer(@CallWindowProcW), nil);
  Result := CallWindowProcW(@Original, hWnd, Msg, WParam, LParam);
  Result := MessageU2WAfter(hWnd, 0, Msg, WParam, LParam, TempWParam, TempLParam, Result, Pointer(@CallWindowProcW), nil);
end;

procedure MessageW2UBefore(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  out OutWParam: WPARAM; out OutLParam: LPARAM; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef);
var
  L: Integer;
begin
  OutWParam := InWParam;
  OutLParam := InLParam;

  case Msg of
    WM_GETTEXT:
      begin
        if Assigned(Proc) then
          L := Proc(hWnd, WM_GETTEXTLENGTH, UnusedWParam, UnusedLParam)
        else
          L := FrameProc(hWnd, hChildWnd, WM_GETTEXTLENGTH, UnusedWParam, UnusedLParam);
        if L > 0 then
        begin
          Inc(L); // wparam needs space for the trailing zero
          GetMem(Pointer(OutLParam), SizeOf(Utf8Byte) * L);
        end
        else
          OutLParam := LPARAM(nil);
        OutWParam := L;
      end;
    LB_GETTEXT:
      begin
        if Assigned(Proc) then
          L := Proc(hWnd, LB_GETTEXTLEN, InWParam, UnusedLParam)
        else
          L := FrameProc(hWnd, hChildWnd, LB_GETTEXTLEN, InWParam, UnusedLParam);
        if L > 0 then
          GetMem(Pointer(OutLParam), SizeOf(Utf8Byte) * (L + 1))
        else
          OutLParam := LPARAM(nil);
      end;
    CB_GETLBTEXT:
      begin
        if Assigned(Proc) then
          L := Proc(hWnd, CB_GETLBTEXTLEN, InWParam, UnusedLParam)
        else
          L := FrameProc(hWnd, hChildWnd, CB_GETLBTEXTLEN, InWParam, UnusedLParam);
        GetMem(Pointer(OutLParam), SizeOf(Utf8Byte) * (L + 1))
      end;
    EM_GETLINE:
      begin
        if InLParam <> 0 then
          L := PWord(InLParam)^
        else
          L := 0;
        if L > 0 then
        begin
          if L < SizeOf(Word) then
            L := SizeOf(Word) div SizeOf(Utf8Byte);
          GetMem(Pointer(OutLParam), SizeOf(Utf8Byte) * L);
          PWord(OutLParam)^ := L;
        end
        else
          OutLParam := LPARAM(nil);
      end;
    WM_MDICREATE:
      begin
        GetMem(Pointer(OutLParam), SizeOf(TMDICreateStructA));
        CopyMemory(Pointer(OutLParam), Pointer(InLParam), SizeOf(TMDICreateStructA));
        PMDICreateStructA(OutLParam).szClass := ConvertToUTF8(PMDICreateStructW(InLParam).szClass);
        PMDICreateStructA(OutLParam).szTitle := ConvertToUTF8(PMDICreateStructW(InLParam).szTitle);
      end;
    WM_SETTEXT, WM_WININICHANGE, WM_DEVMODECHANGE,
    EM_REPLACESEL,
    LB_ADDSTRING, LB_DIR, LB_INSERTSTRING, LB_ADDFILE, LB_FINDSTRING, LB_FINDSTRINGEXACT, LB_SELECTSTRING,
    CB_ADDSTRING, CB_INSERTSTRING, CB_FINDSTRING, CB_DIR, CB_FINDSTRINGEXACT:
      OutLParam := LPARAM(ConvertToUTF8(PWideChar(InLParam)));
    // character fiddling, TODO
    EM_LINELENGTH, EM_LINEINDEX, EM_GETSEL, EM_SETSEL:
      begin
      end;
  else
    // else of case
  end;
end;

function MessageW2UAfter(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  TempWParam: WPARAM; TempLParam: LPARAM; TempResult: LRESULT; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef): LRESULT;
var
  Len: Integer;
  C: PWideChar;
  X: PChar;
begin
  case Msg of
    WM_GETTEXT, LB_GETTEXT, CB_GETLBTEXT:
      begin
        if TempResult >= 0 then
        begin
          C := ConvertFromUTF8(PChar(TempLParam), TempResult, Len);
          if (Len > InWParam) and (Msg = WM_GETTEXT) then
            Len := InWParam;
          TempResult := Len;
          if TempResult <> 0 then
          begin
            if (Msg <> WM_GETTEXT) or (Len < InWParam) then
              Inc(Len); // WParam already includes the #0 but the Result doesn't
            Move(C^, PWideChar(InLParam)^, SizeOf(WideChar) * Len);
          end
          else
            if (InLParam <> 0) and ((Msg <> WM_GETTEXT) or (InWParam > 0)) then
              PWideChar(InLParam)^ := #0;
          FreeMem(C);
        end;
        FreeMem(PChar(TempLParam));
      end;
    EM_GETLINE:
      begin
        if TempResult >= 0 then
        begin
          C := ConvertFromUTF8(PChar(TempLParam), TempResult, Len);
          TempResult := Len;
          if Len > 0 then
            Move(C^, PWideChar(InLParam)^, SizeOf(WideChar) * Len);
          FreeMem(C);
        end;
        FreeMem(PChar(TempLParam));
      end;
    WM_MDICREATE:
      begin
        FreeMemory(PMDICreateStructA(TempLParam).szClass);
        FreeMemory(PMDICreateStructA(TempLParam).szTitle);
        FreeMemory(PMDICreateStructA(TempLParam));
      end;
    WM_SETTEXT, WM_WININICHANGE, WM_DEVMODECHANGE,
    EM_REPLACESEL,
    LB_DIR, LB_ADDFILE, LB_ADDSTRING, LB_INSERTSTRING, LB_FINDSTRING, LB_FINDSTRINGEXACT, LB_SELECTSTRING,
    CB_DIR, CB_ADDSTRING, CB_INSERTSTRING, CB_FINDSTRING, CB_FINDSTRINGEXACT:
      FreeMem(PChar(TempLParam));

    // this can use a little bit of copy/paste reduction
    WM_GETTEXTLENGTH:
      begin
        if TempResult > 0 then
        begin
          // WM_GETTEXTLENGTH returns without trailing #0 WM_GETTEXT wants it
          GetMem(X, SizeOf(Utf8Byte) * (TempResult + 1));
          try
            if Assigned(Proc) then
              Len := Proc(hWnd, WM_GETTEXT, TempResult + 1, LParam(X))
            else
              Len := FrameProc(hWnd, hChildWnd, WM_GETTEXT, TempResult + 1, LParam(X));
            // Len is exclusive #0
            TempResult := ConvertFromUTF8Len(X, Len);
          finally
            FreeMem(X);
          end;
        end;
      end;
    LB_GETTEXTLEN:
      begin
        if TempResult > 0 then
        begin
          GetMem(X, SizeOf(Utf8Byte) * (TempResult + 1));
          try
            if Assigned(Proc) then
              Len := Proc(hWnd, LB_GETTEXT, InWParam, LParam(X))
            else
              Len := FrameProc(hWnd, hChildWnd, LB_GETTEXT, InWParam, LParam(X));
            TempResult := ConvertFromUTF8Len(X, Len);
          finally
            FreeMem(X);
          end;
        end;
      end;
    CB_GETLBTEXTLEN:
      begin
        if TempResult > 0 then
        begin
          GetMem(X, SizeOf(Utf8Byte) * (TempResult + 1));
          try
            if Assigned(Proc) then
              Len := Proc(hWnd, CB_GETLBTEXT, InWParam, LParam(X))
            else
              Len := FrameProc(hWnd, hChildWnd, CB_GETLBTEXT, InWParam, LParam(X));
            TempResult := ConvertFromUTF8Len(X, Len);
          finally
            FreeMem(X);
          end;
        end;
      end;
  end;
  Result := TempResult;
end;

procedure MessageU2WBefore(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  out OutWParam: WPARAM; out OutLParam: LPARAM; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef);
var
  L: Integer;
type
  PMDICreateStructA = ^TMDICreateStructA;
  PMDICreateStructW = ^TMDICreateStructW;
begin
  OutWParam := InWParam;
  OutLParam := InLParam;

  case Msg of
    WM_GETTEXT:
      begin
        if Assigned(FrameProc) then
          L := FrameProc(hWnd, hChildWnd, WM_GETTEXTLENGTH, UnusedWParam, UnusedLParam)
        else
          L := CallWindowProcW(@Proc, hWnd, WM_GETTEXTLENGTH, UnusedWParam, UnusedLParam);
        if L > 0 then
        begin
          Inc(L); // WParam includes the trailing zero
          GetMem(Pointer(OutLParam), SizeOf(WideChar) * L)
        end
        else
          OutLParam := LPARAM(nil);
        OutWParam := L;
      end;
    LB_GETTEXT:
      begin
        if Assigned(FrameProc) then
          L := FrameProc(hWnd, hChildWnd, LB_GETTEXTLEN, InWParam, UnusedLParam)
        else
          L := CallWindowProcW(@Proc, hWnd, LB_GETTEXTLEN, InWParam, UnusedLParam);
        if L > 0 then
          GetMem(Pointer(OutLParam), SizeOf(WideChar) * (L + 1))
        else
          OutLParam := LPARAM(nil);
      end;
    CB_GETLBTEXT:
      begin
        if Assigned(FrameProc) then
          L := FrameProc(hWnd, hChildWnd, CB_GETLBTEXTLEN, InWParam, UnusedLParam)
        else
          L := CallWindowProcW(@Proc, hWnd, CB_GETLBTEXTLEN, InWParam, UnusedLParam);
        GetMem(Pointer(OutLParam), SizeOf(WideChar) * (L + 1))
      end;
    EM_GETLINE:
      begin
        if InLParam <> 0 then
          L := PWord(InLParam)^
        else
          L := 0;
        if L > 0 then
        begin
          if L < SizeOf(Word) then
            L := SizeOf(Word) div SizeOf(WideChar);
          GetMem(Pointer(OutLParam), SizeOf(WideChar) * L);
          PWord(OutLParam)^ := L;
        end
        else
          OutLParam := LPARAM(nil);
      end;
    WM_MDICREATE:
      begin
        GetMem(Pointer(OutLParam), SizeOf(TMDICreateStructW));
        CopyMemory(Pointer(OutLParam), Pointer(InLParam), SizeOf(TMDICreateStructW));
        PMDICreateStructW(OutLParam).szClass := ConvertFromUTF8(PMDICreateStructA(InLParam).szClass);
        PMDICreateStructW(OutLParam).szTitle := ConvertFromUTF8(PMDICreateStructA(InLParam).szTitle);
      end;
    WM_SETTEXT, WM_WININICHANGE, WM_DEVMODECHANGE,
    EM_REPLACESEL,
    LB_DIR, LB_ADDFILE, LB_ADDSTRING, LB_INSERTSTRING, LB_FINDSTRING, LB_FINDSTRINGEXACT, LB_SELECTSTRING,
    CB_DIR, CB_ADDSTRING, CB_INSERTSTRING, CB_FINDSTRING, CB_FINDSTRINGEXACT:
      OutLParam := LPARAM(ConvertFromUTF8(PChar(InLParam)));
    //character fiddling, TODO
    EM_LINELENGTH, EM_LINEINDEX, EM_GETSEL, EM_SETSEL:
      begin

      end;
  end;
end;

function MessageU2WAfter(hWnd, hChildWnd: HWND; Msg: UINT; InWParam: WPARAM; InLParam: LPARAM;
  TempWParam: WPARAM; TempLParam: LPARAM; TempResult: LRESULT; Proc: TStdWndProcDef; FrameProc: TStdFrameWndProcDef): LRESULT;
var
  Len: Integer;
  C: PChar;
  X: PWideChar;
begin
  case Msg of
    WM_GETTEXT, LB_GETTEXT, CB_GETLBTEXT:
      begin
        if TempResult >= 0 then
        begin
          C := ConvertToUTF8(PWideChar(TempLParam), TempResult, Len);
          if (Len > InWParam) and (Msg = WM_GETTEXT) then
            Len := InWParam; // we need more then we allocated so truncate
          TempResult := Len;
          if TempResult <> 0 then
          begin
            if (Msg <> WM_GETTEXT) or (Len < InWParam) then
              Inc(Len); // WParam already includes the #0 but the Result doesn't
            Move(C^, PChar(InLParam)^, SizeOf(Utf8Byte) * Len);
          end
          else
            if (InLParam <> 0) and ((Msg <> WM_GETTEXT) or (InWParam > 0)) then
              PChar(InLParam)^ := #0;
          FreeMem(C);
        end;
        FreeMem(PWideChar(TempLParam));
      end;
    EM_GETLINE:
      begin
        if TempResult >= 0 then
        begin
          C := ConvertToUTF8(PWideChar(TempLParam), TempResult, Len);
          TempResult := Len;
          if Len > 0 then
            Move(C^, PChar(InLParam)^, SizeOf(Utf8Byte) * Len);
          FreeMem(C);
        end;
        FreeMem(PWideChar(TempLParam));
      end;
    WM_MDICREATE:
      begin
        FreeMemory(PMDICreateStructW(TempLParam).szClass);
        FreeMemory(PMDICreateStructW(TempLParam).szTitle);
        FreeMemory(PMDICreateStructW(TempLParam));
      end;
    WM_SETTEXT, WM_WININICHANGE, WM_DEVMODECHANGE,
    EM_REPLACESEL,
    LB_DIR, LB_ADDFILE, LB_ADDSTRING, LB_INSERTSTRING, LB_FINDSTRING, LB_FINDSTRINGEXACT, LB_SELECTSTRING,
    CB_DIR, CB_ADDSTRING, CB_INSERTSTRING, CB_FINDSTRING, CB_FINDSTRINGEXACT:
      FreeMem(PWideChar(TempLParam));
    WM_GETTEXTLENGTH:
      begin
        if TempResult > 0 then
        begin
          // WM_GETTEXTLENGTH returns without trailing #0 WM_GETTEXT wants it
          GetMem(X, SizeOf(WideChar) * (TempResult + 1));
          try
            if Assigned(FrameProc) then
              Len := FrameProc(hWnd, hChildWnd, WM_GETTEXT, TempResult + 1, LParam(X))
            else
              Len := CallWindowProcW(@Proc, hWnd, WM_GETTEXT, TempResult + 1, LParam(X));
            TempResult := ConvertToUTF8Len(X, Len)
          finally
            FreeMem(x);
          end;
        end;
      end;
    LB_GETTEXTLEN:
      begin
        if TempResult > 0 then
        begin
          GetMem(X, SizeOf(WideChar) * (TempResult + 1));
          try
            if Assigned(FrameProc) then
              Len := FrameProc(hWnd, hChildWnd, LB_GETTEXT, InWParam, LParam(X))
            else
              Len := CallWindowProcW(@Proc, hWnd, LB_GETTEXT, InWParam, LParam(X));
            TempResult := ConvertToUTF8Len(X, Len)
          finally
            FreeMem(x);
          end;
        end;
      end;
    CB_GETLBTEXTLEN:
      begin
        if TempResult > 0 then
        begin
          GetMem(X, SizeOf(WideChar) * (TempResult + 1));
          try
            if Assigned(FrameProc) then
              Len := FrameProc(hWnd, hChildWnd, CB_GETLBTEXT, InWParam, LParam(X))
            else
              Len := CallWindowProcW(@Proc, hWnd, CB_GETLBTEXT, InWParam, LParam(X));
            TempResult := ConvertToUTF8Len(X, Len);
          finally
            FreeMem(x);
          end;
        end;
      end;
  end;
  Result := TempResult;
end;

end.

