{**************************************************************************************************}
{                                                                                                  }
{ UTF-8 VCL                                                                                        }
{                                                                                                  }
{ The contents of this file are subject to the Mozilla Public License Version 1.1 (the "License"); }
{ you may not use this file except in compliance with the License. You may obtain a copy of the    }
{ License at http://www.mozilla.org/MPL/                                                           }
{                                                                                                  }
{ Software distributed under the License is distributed on an "AS IS" basis, WITHOUT WARRANTY OF   }
{ ANY KIND, either express or implied. See the License for the specific language governing rights  }
{ and limitations under the License.                                                               }
{                                                                                                  }
{ The Original Code is UTF8VCLUtils.pas                                                            }
{                                                                                                  }
{ The Initial Developer of the Original Code is Bart van der Werf (bwerf, bluelive)                }
{ Portions created by Bart van der Werf are Copyright (C) Bart van der Werf. All rights reserved.  }
{                                                                                                  }
{ Contributor(s):                                                                                  }
{   Andreas Hausladen (ahuser)                                                                     }
{   Robert Marquardt (rom, marquardt)                                                              }
{                                                                                                  }
{**************************************************************************************************}

// Last modified: $Date: 2007/03/31 17:03:48 $

unit UTF8VCLUtils;

{$IFDEF CONDITIONALEXPRESSIONS}
 {$IF CompilerVersion >= 15.0}
  {$WARN UNSAFE_TYPE OFF}
  {$WARN UNSAFE_CODE OFF}
  {$WARN UNSAFE_CAST OFF}
 {$IFEND}
{$ENDIF}

interface

uses
  Classes, Windows, SysUtils;

type
  TStdWndProcDef = function(Window: HWND; Msg, WParam: Longint; LParam: Longint): Longint; stdcall;
  TStdFrameWndProcDef = function(Window, ChildWindow: HWND; Msg, WParam: Longint; LParam: Longint): Longint; stdcall;
  TStdWndProcInterceptDef = function(Original: TStdWndProcDef; Window: HWND;
    Msg, WParam: Longint; LParam: Longint): Longint;

  Utf8Byte = AnsiChar;
  PUtf8Char = PAnsiChar;

function WndProcConvertU2W(Proc: Pointer; U2WIntercept, W2UIntercept: TStdWndProcInterceptDef): Pointer;
function WndProcConvertW2U(Proc: Pointer; U2WIntercept, W2UIntercept: TStdWndProcInterceptDef): Pointer;

function IsU2WWndProc(Potential: Pointer): Boolean;
function IsW2UWndProc(Potential: Pointer): Boolean;

function WrapWndProcU2W(Proc: Pointer; Intercept: TStdWndProcInterceptDef): Pointer;
function WrapWndProcW2U(Proc: Pointer; Intercept: TStdWndProcInterceptDef): Pointer;

procedure InjectJumpHook(OrgProc, NewProc: Pointer);
function ReplaceImport(ModuleBase: Pointer; OrgProc, ToProc: Pointer): Boolean;

function ConvertFromUTF8(lpUTF8: PUtf8Char): PWideChar; overload;
function ConvertFromUTF8(lpUTF8: PUtf8Char; LenIn: Integer; var LenOut: Integer): PWideChar; overload;

function ConvertFromUTF8Len(lpUTF8: PUtf8Char; LenIn: Integer): Integer; // excluding trailing #0

function ConvertToUTF8(lpWide: PWideChar): PUtf8Char; overload;
function ConvertToUTF8(lpWide: PWideChar; LenIn: Integer; var LenOut: Integer): PUtf8Char; overload;
function ConvertToUTF8Len(lpWide: PWideChar; LenIn: Integer): Integer; // excluding trailing #0

function SpecialWndProc(Potential: Cardinal): Boolean;

implementation

const
  SpecialAddress1 = $12345678;
  SpecialAddress2 = $13245768;

// windows also passes around special wndproc values,
// these values are not functionpointers but handles and identifiers
// and need to be treated special
function SpecialWndProc(Potential: Cardinal): Boolean;
begin
  Result := ((Potential and $FFFF0000) = 0) or ((Potential and $FFFF0000) = $FFFF0000);
end;

{$IFNDEF CONDITIONALEXPRESSIONS} // Delphi 5 or lower
const
  RaiseLastOSError: procedure = RaiseLastWin32Error;
{$ENDIF ~CONDITIONALEXPRESSIONS}

type
  PPointer = ^Pointer; // Delphi 5 doesn't declare PPointer

  PWin9xDebugThunk = ^TWin9xDebugThunk;
  TWin9xDebugThunk = packed record
    PUSH: Byte;    // PUSH instruction opcode ($68)
    Addr: Pointer; // The actual address of the DLL routine
    JMP: Byte;     // JMP instruction opcode ($E9)
    Rel: Integer;  // Relative displacement (a Kernel32 address)
  end;

  PAbsoluteIndirectJmp = ^TAbsoluteIndirectJmp;
  TAbsoluteIndirectJmp = packed record
    OpCode: Word;
    Addr: PPointer;
  end;

function IsWin9xDebugThunk(AnAddr: Pointer): Boolean; // extract from JclPeImage.pas
{ -> EAX: AnAddr }
asm
  TEST EAX, EAX
  JZ  @@NoThunk
  CMP BYTE PTR [EAX].TWin9xDebugThunk.PUSH, $68
  JNE @@NoThunk
  CMP BYTE PTR [EAX].TWin9xDebugThunk.JMP, $E9
  JNE @@NoThunk
  XOR EAX, EAX
  MOV AL, 1
  JMP @@exit
@@NoThunk:
  XOR EAX, EAX
@@exit:
end;

function GetActualAddr(Proc: Pointer): Pointer;
begin
  if Proc <> nil then
  begin
    if (SysUtils.Win32Platform <> VER_PLATFORM_WIN32_NT) and IsWin9xDebugThunk(Proc) then
      Proc := PWin9xDebugThunk(Proc).Addr;
    if (PAbsoluteIndirectJmp(Proc).OpCode = $25FF) then // JMP mem32
      Result := PAbsoluteIndirectJmp(Proc).Addr^
    else
      Result := Proc;
  end
  else
    Result := nil;
end;

procedure InjectJumpHook(OrgProc, NewProc: Pointer);
var
  OldProtect: Cardinal;
  Offset: Integer;
begin
  OrgProc := GetActualAddr(OrgProc);
  if OrgProc <> nil then
  begin
    if VirtualProtect(OrgProc, 5, PAGE_EXECUTE_READWRITE, Pointer(OldProtect)) then
    begin
      Offset := Integer(NewProc) - Integer(OrgProc) - 5 { bytes instruction };
      PByte(OrgProc)^ := $E9; // JMP rel32
      Inc(Cardinal(OrgProc));
      PInteger(OrgProc)^ := Offset;
      VirtualProtect(OrgProc, 5, OldProtect, @OldProtect);
      FlushInstructionCache(GetCurrentProcess, OrgProc, 5);
    end;
  end;
end;

type
  TIIDUnion = record
    case Integer of
      0: (Characteristics: DWORD);         // 0 for terminating null import descriptor
      1: (OriginalFirstThunk: DWORD);      // RVA to original unbound IAT (PIMAGE_THUNK_DATA)
  end;

  PIMAGE_IMPORT_DESCRIPTOR = ^IMAGE_IMPORT_DESCRIPTOR;
  _IMAGE_IMPORT_DESCRIPTOR = record
    Union: TIIDUnion;
    TimeDateStamp: DWORD;                  // 0 if not bound,
                                           // -1 if bound, and real date\time stamp
                                           //     in IMAGE_DIRECTORY_ENTRY_BOUND_IMPORT (new BIND)
                                           // O.W. date/time stamp of DLL bound to (Old BIND)

    ForwarderChain: DWORD;                 // -1 if no forwarders
    Name: DWORD;
    FirstThunk: DWORD;                     // RVA to IAT (if bound this IAT has actual addresses)
  end;
  IMAGE_IMPORT_DESCRIPTOR = _IMAGE_IMPORT_DESCRIPTOR;
  TImageImportDescriptor = IMAGE_IMPORT_DESCRIPTOR;
  PImageImportDescriptor = PIMAGE_IMPORT_DESCRIPTOR;

  PIMAGE_THUNK_DATA32 = ^IMAGE_THUNK_DATA32;
  _IMAGE_THUNK_DATA32 = record
    case Integer of
      0: (ForwarderString: DWORD);   // PBYTE
      1: (Function_: DWORD);         // PDWORD
      2: (Ordinal: DWORD);
      3: (AddressOfData: DWORD);     // PIMAGE_IMPORT_BY_NAME
  end;
  IMAGE_THUNK_DATA32 = _IMAGE_THUNK_DATA32;
  TImageThunkData32 = IMAGE_THUNK_DATA32;
  PImageThunkData32 = PIMAGE_THUNK_DATA32;

  IMAGE_THUNK_DATA = IMAGE_THUNK_DATA32;
  PIMAGE_THUNK_DATA = PIMAGE_THUNK_DATA32;
  TImageThunkData = TImageThunkData32;
  PImageThunkData = PImageThunkData32;

function PeMapImgNtHeaders(const BaseAddress: Pointer): PImageNtHeaders;
begin
  Result := nil;
  if IsBadReadPtr(BaseAddress, SizeOf(TImageDosHeader)) then
    Exit;
  if (PImageDosHeader(BaseAddress)^.e_magic <> IMAGE_DOS_SIGNATURE) or
    (PImageDosHeader(BaseAddress)^._lfanew = 0) then
    Exit;
  Result := PImageNtHeaders(DWORD(BaseAddress) + DWORD(PImageDosHeader(BaseAddress)^._lfanew));
  if IsBadReadPtr(Result, SizeOf(TImageNtHeaders)) or
    (Result^.Signature <> IMAGE_NT_SIGNATURE) then
      Result := nil
end;

function ReplaceImport(ModuleBase: Pointer; OrgProc, ToProc: Pointer): Boolean;
var
  ImportThunk: PWin9xDebugThunk;
  NtHeader: PImageNtHeaders;
  ImportDir: TImageDataDirectory;
  ImportDesc: PImageImportDescriptor;
  ImportEntry: PImageThunkData;
  LastProtect: Cardinal;
  WinNT: Boolean;
  FromProc: Pointer;
begin
  Result := False;
  NtHeader := PeMapImgNtHeaders(ModuleBase);
  if (NtHeader = nil) or (ToProc = nil) or (OrgProc = nil) then
    Exit;
  ImportDir := NtHeader.OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT];
  if ImportDir.VirtualAddress = 0 then
    Exit;
  WinNT := Win32Platform = VER_PLATFORM_WIN32_NT;
  ImportDesc := PImageImportDescriptor(DWORD(ModuleBase) + ImportDir.VirtualAddress);
  while ImportDesc^.Name <> 0 do
  begin
    ImportEntry := PImageThunkData(DWORD(ModuleBase) + ImportDesc^.FirstThunk);
    while ImportEntry^.Function_ <> 0 do
    begin
      FromProc := Pointer(ImportEntry^.Function_);
      if not WinNT then
      begin
        ImportThunk := PWin9xDebugThunk(FromProc);
        if IsWin9xDebugThunk(ImportThunk) then
          FromProc := ImportThunk^.Addr;
      end;
      if FromProc = OrgProc then
      begin
        if VirtualProtectEx(GetCurrentProcess, @ImportEntry^.Function_, SizeOf(ToProc),
          PAGE_READWRITE, @LastProtect) then
        begin
          ImportEntry^.Function_ := Cardinal(ToProc);
          VirtualProtectEx(GetCurrentProcess, @ImportEntry^.Function_, SizeOf(ToProc), LastProtect, Pointer(LastProtect));
          Result := True;
        end;
      end;
      Inc(ImportEntry);
    end;
    Inc(ImportDesc);
  end;
end;

function ConvertFromUTF8(lpUTF8: PUtf8Char; LenIn: Integer; var LenOut: Integer): PWideChar;
var
  Len: Integer;
begin
  Result := nil;
  LenOut := 0;
  if not Assigned(lpUTF8) or (LenIn = 0) then
    Exit;
  LenOut := MultiByteToWideChar(CP_UTF8, 0, lpUTF8, LenIn, nil, 0);
  if LenOut = 0 then
    RaiseLastOSError;
  Len := LenOut;
  if lpUTF8[LenIn-1] <> #0 then
    Inc(Len);
  GetMem(Result, SizeOf(WideChar) * (Len));
  Result[Len-1] := #0;
  if MultiByteToWideChar(CP_UTF8, 0, lpUTF8, LenIn, Result, Len) = 0 then
    RaiseLastOSError;
  if Result[LenOut-1] = #0 then
    Dec(LenOut);
end;

function ConvertFromUTF8Len(lpUTF8: PUtf8Char; LenIn: Integer): Integer;
begin
  Result := 0;
  if not Assigned(lpUTF8) or (LenIn = 0) then
    Exit;
  if lpUTF8[LenIn-1] = #0 then
    Dec(LenIn);
  Result := MultiByteToWideChar(CP_UTF8, 0, lpUTF8, LenIn, nil, 0);
  if Result = 0 then
    RaiseLastOSError;
end;

function ConvertFromUTF8(lpUTF8: PUtf8Char): PWideChar;
var
  R, L: Integer;
begin
  Result := nil;
  if not Assigned(lpUTF8) then
    Exit;
  L := lstrlenA(lpUTF8) + 1;
  Result := ConvertFromUTF8(lpUTF8, L, R);
end;

function ConvertToUTF8(lpWide: PWideChar; LenIn: Integer; var LenOut: Integer): PUtf8Char;
var
  Len: Integer;
begin
  Result := nil;
  LenOut := 0;
  if not Assigned(lpWide) then
    Exit;
  if LenIn = 0 then
    Exit;
  LenOut := WideCharToMultiByte(CP_UTF8, 0, lpWide, LenIn, nil, 0, nil, nil);
  if LenOut = 0 then
    RaiseLastOSError;
  Len := LenOut;
  if lpWide[LenIn-1] <> #0 then
    Inc(Len);
  GetMem(Result, SizeOf(Utf8Byte) * (Len));
  Result[Len-1] := #0;
  if WideCharToMultiByte(CP_UTF8, 0, lpWide, LenIn, Result, Len, nil, nil) = 0 then
    RaiseLastOSError;
  if Result[LenOut-1] = #0 then
    Dec(LenOut);
end;

function ConvertToUTF8Len(lpWide: PWideChar; LenIn: Integer): Integer;
begin
  Result := 0;
  if not Assigned(lpWide) or (LenIn = 0) then
    Exit;
  if lpWide[LenIn-1] = #0 then
    Dec(LenIn);

  Result := WideCharToMultiByte(CP_UTF8, 0, lpWide, LenIn, nil, 0, nil, nil);
  if Result = 0 then
    RaiseLastOSError;
end;

function ConvertToUTF8(lpWide: PWideChar): PUtf8Char;
var
  R, L: Integer;
begin
  Result := nil;
  if not Assigned(lpWide) then
    Exit;
  L := lstrlenW(lpWide) + 1;
  Result := ConvertToUTF8(lpWide, L, R);
end;

function StdWndProc(Window: HWND; Msg, WParam: Longint; LParam: Longint): Longint; stdcall;
var
  Jump: TStdWndProcInterceptDef;
begin
  Jump := TStdWndProcInterceptDef(Pointer(SpecialAddress1));
  Result := Jump(TStdWndProcDef(SpecialAddress1), Window, Msg, WParam, LParam);
end;

function MakeWndProcIntercept(Original: TStdWndProcDef; Intercept: TStdWndProcInterceptDef;
  Kind: Integer): TStdWndProcDef;
const
  { This is an assembly representation of StdWndProc }
  StdWndProcStub: array[0..46] of Byte =
   ($55, $8B, $EC, $83, $C4, $F8, $C7,
    $45, $F8, $FE, $FE, $FE, $FE, $8B, $45, $10, $50, $8B, $45, $14, $50, $8B, $4D,
    $0C, $8B, $55, $08, $B8, $FE, $FE, $FE, $FE, $FF, $55, $F8, $89, $45, $FC, $8B,
    $45, $FC, $59, $59, $5D, $C2, $10, $00);
begin
  Result := VirtualAlloc(nil, 256, MEM_COMMIT, PAGE_EXECUTE_READWRITE);
  Move(StdWndProcStub, Pointer(Result)^, Length(StdWndProcStub));
  PPointer(Cardinal(Result) + 9)^ := Pointer(@Intercept);
  PPointer(Cardinal(Result) + 28)^ := Pointer(@Original);
  PInteger(Cardinal(Result) + 64)^ := SpecialAddress2;
  PPointer(Cardinal(Result) + 68)^ := Pointer(@Original);
  PInteger(Cardinal(Result) + 72)^ := Kind;
end;

function ConvertWndProcU2W(Potential: Pointer): Pointer;
begin
  Assert(IsW2UWndProc(Potential));
  Result := PPointer(Cardinal(Potential) + 68)^;
end;

function ConvertWndProcW2U(Potential: Pointer): Pointer;
begin
  Assert(IsU2WWndProc(Potential));
  Result := PPointer(Cardinal(Potential) + 68)^;
end;

function IsConvertableWndProc(Potential: Pointer): Boolean;
var
  MemInfo: TMemoryBasicInformation;
begin
  Result := not SpecialWndProc(Cardinal(Potential));
  if Result then
  begin
    Inc(Cardinal(Potential), 64);
    Result := VirtualQuery(Potential, MemInfo, SizeOf(TMemoryBasicInformation)) <> 0;
    if Result then
      Result := MemInfo.AllocationProtect = PAGE_EXECUTE_READWRITE;
    if Result then
      Result := PInteger(Potential)^ = SpecialAddress2;
  end;
end;

function IsU2WWndProc(Potential: Pointer): Boolean;
begin
  Result := IsConvertableWndProc(Potential);
  if Result then
    Result := PInteger(Cardinal(Potential) + 72)^ = 1;
end;

function IsW2UWndProc(Potential: Pointer): Boolean;
begin
  Result := IsConvertableWndProc(Potential);
  if Result then
    Result := PInteger(Cardinal(Potential) + 72)^ = 2;
end;

function WrapWndProcU2W(Proc: Pointer; Intercept: TStdWndProcInterceptDef): Pointer;
begin
  // Assert(not SpecialWndProc(Cardinal(Proc)));
  Result := @MakeWndProcIntercept(TStdWndProcDef(Proc), Intercept, 1);
  Assert(IsU2WWndProc(Result));
end;

function WrapWndProcW2U(Proc: Pointer; Intercept: TStdWndProcInterceptDef): Pointer;
begin
  Assert(not SpecialWndProc(Cardinal(Proc)));
  Result := @MakeWndProcIntercept(TStdWndProcDef(Proc), Intercept, 2);
  Assert(IsW2UWndProc(Result));
end;

function WndProcConvertU2W(Proc: Pointer; U2WIntercept, W2UIntercept: TStdWndProcInterceptDef): Pointer;
begin
  if not Assigned(Proc) then
  begin
    Result := nil;
    Exit;
  end;
  Assert(not IsU2WWndProc(Proc));
  if IsW2UWndProc(Proc) then
    Result := ConvertWndProcU2W(Proc)
  else
    Result := WrapWndProcU2W(Proc, U2WIntercept);
  Assert(not IsW2UWndProc(Result));
end;

function WndProcConvertW2U(Proc: Pointer; U2WIntercept, W2UIntercept: TStdWndProcInterceptDef): Pointer;
begin
  if not Assigned(Proc) then
  begin
    Result := nil;
    Exit;
  end;
  Assert(not IsW2UWndProc(Proc));
  if IsU2WWndProc(Proc) then
    Result := ConvertWndProcW2U(Proc)
  else
    Result := WrapWndProcW2U(Proc, W2UIntercept);
  Assert(not IsU2WWndProc(Result));
end;

end.

